# -*- coding: utf-8 -*-

# Well known methods
GET = 'GET'
HEAD = 'HEAD'
POST = 'POST'
PUT = 'PUT'
PATCH = 'PATCH'
DELETE = 'DELETE'
OPTIONS = 'OPTIONS'

# Type of path
PATH_TYPE_BASE = 'base'
PATH_TYPE_DETAIL = 'detail'

PATH_TYPE_COLLECTION = 'collection'
PATH_TYPE_RESOURCE = 'resource'
